import argparse
from contextlib import contextmanager, ExitStack
from glob import glob
import os
import sys
import time

import migra
import psycopg2
from psycopg2 import sql
import sqlalchemy

from .config import Config

TUSKER_COMMENT = (
    'CREATED BY TUSKER - If this table is left behind tusker probably '
    'crashed and was not able to clean up after itself. Either try '
    'running `tusker clean` or remove this database manually.'
)


try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata
try:
    __version__ = importlib_metadata.version(__name__)
except:
    __version__ = 'unknown'


def execute_sql_file(cursor, filename):
    with open(filename) as fh:
        sql = fh.read()
        sql = sql.strip()
        if sql:
            sql = sqlalchemy.text(sql)
            cursor.execute(sql)


class Tusker:

    def __init__(self, config: Config, verbose=False):
        self.config = config
        self.verbose = verbose
        self.conn = self._connect('template1')
        self.conn.autocommit = True

    def _connect(self, name):
        args = self.config.database.args(dbname='template1')
        return psycopg2.connect(**args)

    def log(self, text):
        if self.verbose:
            print(text, file=sys.stderr)

    @contextmanager
    def createengine(self, dbname=None):
        override = {'dbname': dbname} if dbname else {}
        engine = sqlalchemy.create_engine(
            'postgresql://',
            connect_args=self.config.database.args(**override)
        )
        try:
            yield engine
        finally:
            engine.dispose()

    @contextmanager
    def createdb(self, suffix):
        cursor = self.conn.cursor()
        now = int(time.time())
        dbname = '{}_{}_{}'.format(
            self.config.database.args()['dbname'],
            now,
            suffix
        )
        cursor.execute(sql.SQL('CREATE DATABASE {}').format(
            sql.Identifier(dbname)
        ))
        cursor.execute(sql.SQL('COMMENT ON DATABASE {} IS {}').format(
            sql.Identifier(dbname),
            sql.Literal(TUSKER_COMMENT)
        ))
        try:
            with self.createengine(dbname) as engine:
                yield engine
        finally:
            cursor.execute(sql.SQL('DROP DATABASE {}').format(
                sql.Identifier(dbname)
            ))

    @contextmanager
    def mgr_schema(self):
        with self.createdb('schema') as schema_engine:
            with schema_engine.begin() as schema_cursor:
                self.log('Creating original schema...')
                for filename in sorted(glob(self.config.schema.filename)):
                    self.log('- {}'.format(filename))
                    execute_sql_file(schema_cursor, filename)
            yield schema_engine

    @contextmanager
    def mgr_migrations(self):
        with self.createdb('migrations') as migrations_engine:
            with migrations_engine.begin() as migrations_cursor:
                self.log('Creating migrated schema...')
                for filename in self._get_migrations():
                    self.log('- {}'.format(filename))
                    execute_sql_file(migrations_cursor, filename)
            yield migrations_engine

    @contextmanager
    def mgr_database(self):
        with self.createengine() as database_engine:
            with database_engine.begin() as database_cursor:
                self.log('Observing database schema...')
            yield database_engine

    def mgr(self, name):
        return getattr(self, 'mgr_{}'.format(name))()

    def diff(self, source, target, with_privileges=False):
        self.log('Creating databases...')
        with self.mgr(source) as source, self.mgr(target) as target:
            self.log('Diffing...')
            migration = migra.Migration(
                source,
                target,
                self.config.database.schema
            )
            migration.set_safety(False)
            migration.add_all_changes(privileges=with_privileges)
            return migration.sql

    def check(self, backends, with_privileges=False):
        with ExitStack() as stack:
            managers = [(name, stack.enter_context(self.mgr(name)))
                        for name in backends]
            for i in range(len(managers)-1):
                source, target = (managers[i], managers[i+1])
                self.log('Diffing {} against {}...'.format(
                    source[0],
                    target[0]
                ))
                migration = migra.Migration(
                    source[1],
                    target[1],
                    schema=self.config.database.schema
                )
                migration.set_safety(False)
                migration.add_all_changes(privileges=with_privileges)
                if migration.sql:
                    return (source[0], target[0])
        return None

    def clean(self):
        cursor = self.conn.cursor()
        try:
            cursor.execute('''
                SELECT db.datname
                FROM pg_database db
                JOIN pg_shdescription dsc ON dsc.objoid = db.oid
                WHERE dsc.description = %s;
            ''', (TUSKER_COMMENT,))
            rows = cursor.fetchall()
            for row in rows:
                dbname = row[0]
                self.log('Dropping {} ...'.format(dbname))
                cursor.execute(sql.SQL('DROP DATABASE {}').format(
                    sql.Identifier(dbname)
                ))
        finally:
            cursor.close()

    def _get_migrations(self):
        migrations = []
        if self.config.migrations.filename:
            migrations = glob(self.config.migrations.filename)
        else:
            migrations = [
                os.path.join(self.config.migrations.directory, filename)
                for filename in os.listdir(self.config.migrations.directory)
                if filename.endswith('.sql')
            ]
        return sorted(migrations)

def cmd_diff(args, cfg: Config):
    tusker = Tusker(cfg, args.verbose)
    source = args.source
    target = args.target
    if args.reverse:
        source, target = target, source
    sql = tusker.diff(source, target, with_privileges=args.with_privileges)
    print(sql, end='')


def cmd_check(args, cfg: Config):
    backends = args.backends
    if 'all' in backends:
        backends = ['migrations', 'schema', 'database']
    tusker = Tusker(cfg, args.verbose)
    diff = tusker.check(backends, with_privileges=args.with_privileges)
    if diff:
        print('Schemas differ: {} != {}'.format(diff[0], diff[1]))
        print('Run `tusker diff` to see the differences')
        sys.exit(1)
    else:
        print('Schemas are identical')
        sys.exit(0)


def cmd_clean(args, cfg: Config):
    tusker = Tusker(cfg, args.verbose)
    tusker.clean()


BACKEND_CHOICES = ['migrations', 'schema', 'database']


class ValidateBackends(argparse.Action):
    def __call__(self, parser, args, values, option_string=None):
        if 'all' in values:
            values = BACKEND_CHOICES
        else:
            if len(values) <= 1:
                choices = ', '.join(map(repr, BACKEND_CHOICES))
                raise argparse.ArgumentError(
                    self,
                    (
                        'at least two backends are required to perform '
                        'the check (choose from {choices}) or pass \'all\' '
                        'on its own.'
                    )
                )
            for value in values:
                if value not in BACKEND_CHOICES:
                    choices = ', '.join(map(repr, BACKEND_CHOICES + ['all']))
                    msg = 'invalid choice: {!r} (choose from {})'.format(
                        value,
                        choices
                    )
                    raise argparse.ArgumentError(self, msg)
        setattr(args, self.dest, values)


def main():
    parser = argparse.ArgumentParser(
        description='Generate a database migration.')
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(__version__))
    parser.add_argument(
        '--verbose',
        help='enable verbose output',
        action='store_true',
        default=False)
    parser.add_argument(
        '--config', '-c',
        help='the configuration file. Default: tusker.toml',
        default='tusker.toml')
    subparsers = parser.add_subparsers(
        dest='command',
        required=True)
    parser_diff = subparsers.add_parser(
        'diff',
        help='show differences between two schemas',
        description='''
            This command calculates the difference between two database schemas.
            The from- and to-parameter accept one of the following backends:
            migrations, schema, database
        ''')
    parser_diff.add_argument(
        'source',
        metavar='from',
        nargs='?',
        help='from-backend for the diff operation. Default: migrations',
        choices=BACKEND_CHOICES,
        default='migrations')
    parser_diff.add_argument(
        'target',
        metavar='to',
        nargs='?',
        help='to-backend for the diff operation. Default: schema',
        choices=BACKEND_CHOICES,
        default='schema')
    parser_diff.add_argument(
        '--reverse', '-r',
        help='swaps the "from" and "to" arguments creating a reverse diff',
        action='store_true')
    parser_diff.add_argument(
        '--with-privileges',
        help='also output privilege differences (ie. grant/revoke statements)',
        action='store_true')
    parser_diff.set_defaults(func=cmd_diff)
    parser_check = subparsers.add_parser(
        'check',
        help='check for differences between schemas',
        description='''
            This command checks for differences between two or more schemas.
            Exit code 0 means that the schemas are all in sync. Otherwise the
            exit code 1 is used. This is useful for continuous integration checks.
        ''')
    parser_check.set_defaults(func=cmd_check)
    parser_check.add_argument(
        'backends',
        help=(
            'at least two backends are required to diff against each other '
            '(choose from {}). You can also pass \'all\' on its own to diff '
            'all backends against each other.'
        ).format(
            ', '.join(map(repr, BACKEND_CHOICES))
        ),
        metavar='backend',
        nargs='*',
        default=['migrations', 'schema'],
        action=ValidateBackends
    )
    parser_check.add_argument(
        '--with-privileges',
        help='also output privilege differences (ie. grant/revoke statements)',
        action='store_true')
    parser_clean = subparsers.add_parser(
        'clean',
        help='clean up left over *_migrations or *_schema tables')
    parser_clean.set_defaults(func=cmd_clean)
    args = parser.parse_args()
    if hasattr(args, 'from') and hasattr(args, 'target') and args.source == args.target:
        parser.error('source and target must not be identical')
    cfg = Config(args.config)
    args.func(args, cfg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tusker']

package_data = \
{'': ['*']}

install_requires = \
['migra>=3.0.1621480950,<4.0.0',
 'psycopg2-binary>=2.9.1,<3.0.0',
 'sqlalchemy>=1.4.25,<2.0.0',
 'tomlkit>=0.7.2,<0.8.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

entry_points = \
{'console_scripts': ['tusker = tusker:main']}

setup_kwargs = {
    'name': 'tusker',
    'version': '0.4.1',
    'description': 'A PostgreSQL specific migration tool',
    'long_description': '# Tusker\n\n[![GitHub](https://img.shields.io/github/license/bikeshedder/tusker?label=License&logoColor=white&style=for-the-badge)](https://github.com/bikeshedder/tusker/blob/master/LICENSE)\n&nbsp;\n[![PyPI](https://img.shields.io/pypi/v/tusker?label=PyPI&logo=pypi&logoColor=white&style=for-the-badge)](https://pypi.org/project/tusker)\n\nA PostgreSQL specific migration tool\n\n## Elevator pitch\n\nDo you want to write your database schema directly as SQL\nwhich is understood by PostgreSQL?\n\nDo you want to be able to make changes to this schema and\ngenerate the SQL which is required to migrate between the\nold and new schema version?\n\nTusker does exactly this.\n\n## Installation\n\n```shell\npipx install tusker\n```\n\nNow you should be able to run tusker. Give it a try:\n\n```shell\ntusker --help\n```\n\n## Getting started\n\nOnce tusker is installed create a new file called `schema.sql`:\n\n```sql\nCREATE TABLE fruit (\n    id BIGINT GENERATED BY DEFAULT AS IDENTITY,\n    name TEXT NOT NULL UNIQUE\n);\n```\n\nTusker also needs an empty `migrations` directory:\n\n```shell\nmkdir migrations\n```\n\nNow you should be able to create your first migration:\n\n```\ntusker diff\n```\n\nThe migration is printed to the console and all you need to do is\ncopy and paste the output into a new file in the migrations directory.\nAlternatively you can also pipe the output of `tusker diff` into the\ntarget file:\n\n```\ntusker diff > migrations/0001_initial.sql\n```\n\nAfter that check that your `schema.sql` and your `migrations` are in sync:\n\n```\ntusker diff\n```\n\nThis should give you an empty output. This means that there is no difference\nbetween applying the migrations in order and the target schema.\n\nAlternatively you can run the check command:\n\n```\ntusker check\n```\n\nIf you want to change the schema in the future simply change the `schema.sql`\nand run `tusker diff` to create the migration for you.\n\nGive it a try and change the `schema.sql`:\n\n```sql\nCREATE TABLE fruit (\n    id BIGINT GENERATED BY DEFAULT AS IDENTITY,\n    name TEXT NOT NULL UNIQUE,\n    color TEXT NOT NULL DEFAULT \'\'\n);\n```\n\nCreate a new migration:\n\n```\ntusker diff > migrations/0002_fruit_color.sql\n```\n\n**Congratulations! You are now using SQL to write your migrations. You are no longer limited by a 3rd party data definition language or an object relational wrapper.**\n\n## Configuration\n\nIn order to run tusker you do not need a configuration file. The following\ndefaults are assumed:\n\n- The file containing your database schema is called `schema.sql`\n- The directory containing the migrations is called `migrations`\n- Your current user can connect to the database using a unix\n  domain socket without a password.\n\nYou can also create a configuration file called `tusker.toml`. The default\nconfiguration looks like that:\n\n```toml\n[schema]\nfilename = "schema.sql"\n\n[migrations]\nfilename = "migrations/*.sql"\n\n[database]\n#host = ""\n#port = 5432\n#user = ""\n#password = ""\ndbname = "tusker"\n```\n\nInstead of the exploded form of `host`, `port`, etc. it\nis also possible to pass a connection URL:\n\n```toml\n[schema]\nfilename = "schema.sql"\n\n[migrations]\nfilename = "migrations/*.sql"\n\n[database]\nurl = "postgresql:///my_awesome_db"\n```\n\n## How can I use the generated SQL files?\n\nThe resulting SQL files can either be applied to the database by hand\nor by using one of the many great tools and libraries which support\napplying SQL files in order.\n\nSome recommendations are:\n\n- NodeJS: [marv](https://www.npmjs.com/package/marv)\n- Rust: [refinery](https://crates.io/crates/refinery)\n\n## How does it work?\n\nUpon startup `tusker` reads all files from the `migrations` directory\nand runs them on an empty database. Another empty database is created\nand the target schema is created. Then those two schemas are\ndiffed using the excellent [migra](https://pypi.org/project/migra/)\ntool and the output printed to the console.\n\n## FAQ\n\n### Is it possible to split the schema into multiple files?\n\nYes. This feature has been added in 0.3. You can now use `glob` patterns as\npart of the `schema.filename` setting. e.g.:\n\n```toml\n[schema]\nfilename = "schema/*.sql"\n```\n\n### Is it possible to diff the schema and/or migrations against an existing database?\n\nYes. This feature has been added in 0.2. You can pass a `from` and `to`\nargument to the `tusker diff` command. Check the output of `tusker diff --help` for\nmore details.\n\n### Tusker printed an error and left the temporary databases behind. How can I remove them?\n\nRun `tusker clean`. This will remove all databases which were created\nby previous runs of tusker. Tusker only removes databases which are\nmarked with a `CREATED BY TUSKER` comment.\n\n### What does the `dbname` setting in `tusker.toml` mean?\n\nWhen diffing against a ready migrated database this database name is used. This\ncommand will print out the difference between the current database schema and\nthe target schema:\n\n```shell\ntusker diff database\n```\n\nTusker also needs to create temporary databases when diffing against the `schema`\nand/or `migrations`. The two databases are called `{dbname}_{timestamp}_schema`\nand `{dbname}_{timestamp}_migrations`.\n',
    'author': 'Michael P. Jung',
    'author_email': 'michael.jung@terreon.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bikeshedder/tusker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import os

from psycopg2.extensions import parse_dsn
from tomlkit.toml_file import TOMLFile


class Config:

    def __init__(self, filename=None):
        env = os.environ
        filename = filename or 'tusker.toml'
        toml = TOMLFile(filename)
        try:
            data = toml.read()
        except FileNotFoundError:
            data = {}
        # time to validate some configuration variables
        data.setdefault('database', {'dbname': 'tusker'})
        data.setdefault('schema', {'filename': 'schema.sql'})
        data.setdefault('migrations', {'directory': 'migrations'})
        self.schema = SchemaConfig(data['schema'])
        self.migrations = MigrationsConfig(data['migrations'])
        self.database = DatabaseConfig(data['database'])

    def __str__(self):
        return 'Config(schema={}, migrations={}, database={})'.format(self.schema, self.migrations, self.database)


class ConfigReader:

    def __init__(self, data, path):
        self.data = data
        self.path = path

    def get(self, name, type, required=False):
        if name not in self.data:
            if required:
                raise ConfigError.missing('{}.{}'.format(path, name))
            else:
                return None
        value = self.data[name]
        if not isinstance(value, type):
            raise ConfigError.invalid(name, 'Not of type {}'.format(type))
        return value


class SchemaConfig:

    def __init__(self, data):
        data = ConfigReader(data, 'schema')
        self.filename = data.get('filename', str) or 'schema.sql'

    def __str__(self):
        return 'SchemaConfig({!r})'.format(self.__dict__)


class MigrationsConfig:

    def __init__(self, data):
        data = ConfigReader(data, 'migrations')
        self.directory = data.get('directory', str, False)
        self.filename = data.get('filename', str, False)
        if not self.directory and not self.filename:
            self.directory = 'migrations'
        elif self.directory and self.filename:
            raise ConfigError.invalid(
                'migrations',
                'directory and filename parameters are mutually exclusive',
            )

    def __str__(self):
        return 'MigrationsConfig({!r})'.format(self.__dict__)


class DatabaseConfig:

    def __init__(self, data):
        data = ConfigReader(data, 'database')
        self.url = data.get('url', str)
        self.host = data.get('host', str)
        self.port = data.get('port', int)
        self.dbname = data.get('dbname', str)
        self.user = data.get('user', str)
        self.password = data.get('password', str)
        self.schema = data.get('schema', str)

    def __str__(self):
        return 'DatabaseConfig({!r})'.format(self.__dict__)

    def args(self, **override):
        if self.url:
            args = parse_dsn(self.url)
        else:
            args = {}
        for k in ['host', 'port', 'dbname', 'user', 'password']:
            v = getattr(self, k)
            if v is not None:
                args[k] = v
        if not args['dbname']:
            args['dbname'] = 'tusker'
        args.update(override)
        return args


class ConfigError(RuntimeError):

    @classmethod
    def missing(cls, name):
        return cls('Missing configuration: {}'.format(name))

    @classmethod
    def invalid(cls, name, reason):
        return cls('Invalid configuration: {}, {}'.format(name, reason))
